"""Parameter wrappers."""
from __future__ import annotations

from collections.abc import Iterable
from enum import Enum
from functools import partial
from itertools import product
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    List,
    Sequence,
    Tuple,
    TypedDict,
    cast,
    overload,
)

if TYPE_CHECKING:
    from .problem import Problem, ProblemOutputType, ProblemParamSpec


# pylint: disable=C0103
class AgaReservedKeywords(Enum):
    """Reserved keywords for aga."""

    aga_expect = "aga_expect"
    aga_expect_stdout = "aga_expect_stdout"
    aga_hidden = "aga_hidden"
    aga_name = "aga_name"
    aga_weight = "aga_weight"
    aga_value = "aga_value"
    aga_extra_credit = "aga_extra_credit"
    aga_override_check = "aga_override_check"
    aga_override_test = "aga_override_test"
    aga_description = "aga_description"
    aga_is_pipeline = "aga_is_pipeline"


DEFAULT_AGA_RESERVED_VALUES = {
    "aga_expect": None,
    "aga_expect_stdout": None,
    "aga_hidden": False,
    "aga_name": None,
    "aga_weight": 1,
    "aga_value": 0.0,
    "aga_extra_credit": 0.0,
    "aga_override_check": None,
    "aga_override_test": None,
    "aga_description": None,
    "aga_is_pipeline": False,
}


class AgaKeywordDictType(TypedDict):
    """Aga keyword arguments type."""

    aga_expect: None | Any
    aga_expect_stdout: None | str
    aga_hidden: bool
    aga_name: None | str
    aga_weight: int
    aga_value: float
    aga_extra_credit: float
    aga_override_check: None | Callable[..., bool]
    aga_override_test: None | Callable[..., Any]
    aga_description: None | str
    aga_is_pipeline: bool


__all__ = [
    "test_case",
    "param",
    "test_cases",
    "test_cases_params",
    "test_cases_zip",
    "test_cases_product",
    "test_cases_singular_params",
    "AgaReservedKeywords",
    "AgaKeywordContainer",
]


def _check_default_values() -> None:
    """Raise an error if `kwd` is reserved."""
    assert len(AgaReservedKeywords) == len(DEFAULT_AGA_RESERVED_VALUES)
    for kwd in AgaReservedKeywords:
        assert kwd.value in DEFAULT_AGA_RESERVED_VALUES


_check_default_values()


class AgaKeywordContainer:
    """A container for aga_* keyword arguments."""

    def __init__(self, **kwargs: Any):
        self.aga_kwargs: AgaKeywordDictType = cast(AgaKeywordDictType, kwargs)

    @property
    def aga_kwargs(self) -> AgaKeywordDictType:
        """Return the aga_* keyword arguments of the test."""
        return self._aga_kwargs

    @aga_kwargs.setter
    def aga_kwargs(self, kwargs: AgaKeywordDictType) -> None:
        """Set the aga_* keyword arguments of the test."""
        self._aga_kwargs = kwargs
        self.ensure_aga_kwargs()

    def ensure_aga_kwargs(self) -> AgaKeywordContainer:
        """Ensure that the aga_* keywords are handled correct."""
        for k in self.aga_kwargs:
            try:
                AgaReservedKeywords(k)
            except ValueError as e:
                raise ValueError(f'invalid kwargs "{k}" in a test param') from e
        return self

    def update_aga_kwargs(self, **kwargs: Any) -> AgaKeywordContainer:
        """Update the keyword arguments to be passed to the functions under test."""
        self.aga_kwargs.update(kwargs)  # type: ignore
        self.ensure_aga_kwargs()
        return self

    def ensure_default_aga_values(self) -> AgaKeywordContainer:
        """Ensure that the aga_* keywords all have default."""
        self.aga_kwargs = cast(
            AgaKeywordDictType, {**DEFAULT_AGA_RESERVED_VALUES, **self.aga_kwargs}
        )
        return self

    @property
    def description(self) -> str | None:
        """Get the description of the test case."""
        return self.aga_kwargs[AgaReservedKeywords.aga_description.value]

    @description.setter
    def description(self, desc: str | None) -> None:
        """Set the description of the test case."""
        self.aga_kwargs[AgaReservedKeywords.aga_description.value] = desc

    @property
    def name(self) -> str | None:
        """Get the name of the test case."""
        return self.aga_kwargs[AgaReservedKeywords.aga_name.value]

    @name.setter
    def name(self, name: str | None) -> None:
        """Set the name of the test case."""
        self.aga_kwargs[AgaReservedKeywords.aga_name.value] = name

    @property
    def override_test(self) -> Callable[..., Any] | None:
        """Get the override_test aga_override_test of the test case."""
        return self.aga_kwargs[AgaReservedKeywords.aga_override_test.value]

    @property
    def override_check(self) -> Callable[..., Any] | None:
        """Get the override_check aga_override_check of the test case."""
        return self.aga_kwargs[AgaReservedKeywords.aga_override_check.value]

    @property
    def is_pipeline(self) -> bool:
        """Get the is_pipeline aga_is_pipeline of the test case."""
        return self.aga_kwargs[AgaReservedKeywords.aga_is_pipeline.value]

    @property
    def weight(self) -> int:
        """Get the weight aga_weight of the test case."""
        return self.aga_kwargs[AgaReservedKeywords.aga_weight.value]

    @property
    def value(self) -> float:
        """Get the value aga_value of the test case."""
        return self.aga_kwargs[AgaReservedKeywords.aga_value.value]

    @property
    def extra_credit(self) -> float:
        """Get the extra credit aga_extra_credit of the test case."""
        return self.aga_kwargs[AgaReservedKeywords.aga_extra_credit.value]

    @property
    def hidden(self) -> bool:
        """Get the hidden aga_hidden of the test case."""
        return self.aga_kwargs[AgaReservedKeywords.aga_hidden.value]

    @property
    def expect(self) -> Any:
        """Get the expected aga_expect of the test case."""
        return self.aga_kwargs[AgaReservedKeywords.aga_expect.value]

    @property
    def expect_stdout(self) -> str | None:
        """Get the expected aga_expect_stdout of the test case."""
        return self.aga_kwargs[AgaReservedKeywords.aga_expect_stdout.value]

    def aga_kwargs_repr(self, sep: str = ",") -> str:
        """Return a string representation of the test's aga_* keyword arguments."""
        return sep.join(k + "=" + repr(v) for k, v in self.aga_kwargs.items())


class _TestParam(AgaKeywordContainer):
    __slots__ = ["_args", "_kwargs", "_aga_kwargs"]

    pipeline: ClassVar[partial[_TestParam]]

    @overload
    def __init__(
        self,
        *args: Any,
        aga_expect: Any = None,
        aga_expect_stdout: str | Sequence[str] | None = None,
        aga_hidden: bool = False,
        aga_name: str | None = None,
        aga_description: str | None = None,
        aga_weight: int = 1,
        aga_value: float = 0.0,
        aga_extra_credit: float = 0.0,
        aga_override_check: Callable[..., Any] | None = None,
        aga_override_test: Callable[..., Any] | None = None,
        aga_is_pipeline: bool = False,
        **kwargs: Any,
    ) -> None:
        ...

    @overload
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        ...

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        r"""Declare a specific test case/param for some problem.

        Parameters
        ----------
        args :
            The arguments to be passed to the functions under test.
        aga_expect : Optional[T]
            If aga_expect is None, the inputs will be tested against the wrapped
            function, the "golden solution" to the problem. If aga_expect is
            specified, the inputs will double as a test _of_ the golden solution;
            to successfully produce the problem grader, the golden solution must
            return aga_expect from the given input.
        aga_expect_stdout : Optional[str | Sequence[str]]
            If aga_expect_stdout is specified, the golden solution's stdout will be
            checked against the given value(s). If aga_expect_stdout is a string, it
            will be compared against the golden solution's stdout. If
            aga_expect_stdout is a sequence of strings, the golden solution's stdout
            will be split on newlines (with '\n' removed) and the resulting list will
            be compared against the given sequence. If aga_expect_stdout is None, the
            golden solution's stdout will not be checked.
        aga_hidden : bool
            If True, hide the problem from students on supported frontends.
        aga_name : Optional[str]
            The test case's name. If `None`, defaults to "Test {inputs}", where
            {inputs} is a comma-separated list of args and kwargs.
        aga_description: Optional[str]
            The detailed description for the test case. It will be displayed under the
            test case name and thus supports longer descriptions.
        aga_weight : int
            The test case's relative weight to the group's score. See :ref:`Determining
            Score` for details.
        aga_value : float
            The test case's absolute score. See :ref:`Determining Score` for details.
        aga_extra_credit : float
            The test case's absolute extra credit score. See :ref:`Determining Score`
            for details.
        aga_override_check : Callable[[TestCase, Output, T], None] | None
            A function which overrides the equality assertions made by the library. See
            :ref:`Overriding the Equality Check` for more.
        aga_override_test : Callable[[TestCase, Callable[T], Callable[T]], None] | None
            A function which overrides the entire test behavior of the library. See
            :ref:`Overriding the Entire Test` for more.
        aga_is_pipeline: bool
            If True, the test case will be run through as a pipeline.
        kwargs :
            Keyword arguments to be passed to the functions under test. Any keyword
            starting with aga\_ is reserved.

        Returns
        -------
        Callable[[Problem[T]], Problem[T]]
            A decorator which adds the test case to a problem.
        """
        super().__init__(
            **{
                k.value: kwargs.pop(k.value)
                for k in AgaReservedKeywords
                if k.value in kwargs
            }
        )
        self.kwargs = kwargs
        self.args = args

    @property
    def args(self) -> Tuple[Any, ...]:
        """Return the arguments to be passed to the functions under test."""
        return self._args

    @args.setter
    def args(self, args: Tuple[Any, ...]) -> None:
        """Set the arguments to be passed to the functions under test."""
        self._args = args

    @property
    def kwargs(self) -> Dict[str, Any]:
        """Return the keyword arguments to be passed to the functions under test."""
        return self._kwargs

    @kwargs.setter
    def kwargs(self, kwargs: Dict[str, Any]) -> None:
        """Set the keyword arguments to be passed to the functions under test."""
        self._kwargs = kwargs
        self.ensure_valid_kwargs()

    def ensure_valid_kwargs(self) -> _TestParam:
        """Ensure that the aga_* keywords are handled correct."""
        for k in self.kwargs:
            if k.startswith("aga_"):
                raise ValueError(
                    f'aga keyword "{k}" should not be in kwargs of a test param'
                )
        return self

    def args_repr(self, sep: str = ",") -> str:
        """Return a string representation of the test's arguments."""
        return sep.join(repr(x) for x in self.args)

    def kwargs_repr(self, sep: str = ",") -> str:
        """Return appropriate string representation of the test's keyword arguments."""
        # we use k instead of repr(k) so we don't get quotes around it
        return sep.join(k + "=" + repr(v) for k, v in self.kwargs.items())

    def sep_repr(self, sep: str = ",") -> str:
        """Return sep if both exist, "" otherwise."""
        return self.args and self.kwargs and sep or ""

    def generate_test_case(
        self, prob: Problem[ProblemParamSpec, ProblemOutputType]
    ) -> Problem[ProblemParamSpec, ProblemOutputType]:
        """Generate a test case for the given problem."""
        self.ensure_default_aga_values()
        self.ensure_valid_kwargs().ensure_aga_kwargs()  # type: ignore

        prob.add_test_case(param=self)

        return prob

    def __call__(
        self, prob: Problem[ProblemParamSpec, ProblemOutputType]
    ) -> Problem[ProblemParamSpec, ProblemOutputType]:
        """Add the test case to the given as a decorator."""
        return self.generate_test_case(prob)

    def __str__(self) -> str:
        """Return a string representation of the test case."""
        return f"param({self.args_repr()}, {self.kwargs_repr()})"

    def __repr__(self) -> str:
        """Return a string representation of the test case."""
        return (
            f"param({self.args_repr()}, {self.kwargs_repr()}, {self.aga_kwargs_repr()})"
        )


_TestParam.pipeline = partial(
    _TestParam, aga_is_pipeline=True
)  # pylint: disable=invalid-name
param = _TestParam  # pylint: disable=invalid-name
test_case = _TestParam  # pylint: disable=invalid-name


class _TestParams:
    """A class to store the parameters for a test."""

    __slots__ = ["final_params"]

    params: ClassVar[partial[_TestParams]]
    zip: ClassVar[partial[_TestParams]]
    product: ClassVar[partial[_TestParams]]
    singular_params: ClassVar[partial[_TestParams]]

    # pylint: disable=too-many-locals
    @overload
    def __init__(
        self,
        *args: Any,
        aga_expect: Any = None,
        aga_expect_stdout: str | Sequence[str] | None = None,
        aga_hidden: bool = False,
        aga_name: str | None = None,
        aga_description: str | None = None,
        aga_weight: int = 1,
        aga_value: float = 0.0,
        aga_extra_credit: float = 0.0,
        aga_override_check: Callable[..., Any] | None = None,
        aga_override_test: Callable[..., Any] | None = None,
        aga_is_pipeline: bool = False,
        aga_product: bool = False,
        aga_zip: bool = False,
        aga_params: bool = False,
        aga_singular_params: bool = False,
        **kwargs: Any,
    ):
        ...

    @overload
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        ...

    def __init__(
        self,
        *args: Iterable[Any] | Any,
        aga_product: bool = False,
        aga_zip: bool = False,
        aga_params: bool = False,
        aga_singular_params: bool = False,
        **kwargs: Any,
    ) -> None:
        r"""Generate many test cases programmatically, from generators of inputs.

        Parameters
        ----------
        args :
            Generators for arguments to be passed to the functions under test.
        aga_product : bool
            Whether to take a cartesian product of the generators (creating one test
            case for each set of inputs in the product), or to zip them (iterate
            through each generator in sequence). Default `False`.
        aga_zip : bool
            Whether to zip the input generators. Default `False`.
        aga_params : bool
            Whether to treat the input generators as a single iterable of sets of
            parameters. Default `False`.
        aga_singular_params : bool
            Whether to treat the input generators as a single iterable of single params.
            Default `False`.
        kwargs :
            `aga_` keywords have their meaning inherited from `test_case`, and are
            applied to each test case generated by this function. Singleton value and
            Sequence values are supported, i.e., `aga_hidden=True` and
            `aga_expect = [1, 2, 3]` are both valid. When passing in a Sequence object,
            the length of the sequence must match the length of the test cases.
            Generators for keyword arguments to be passed to the functions under test.
            Any keyword starting with aga\_ is reserved.

        Returns
        -------
        Callable[[Problem[T]], Problem[T]]
            A decorator which adds the test cases to a problem.
        """
        if aga_params + aga_zip + aga_product + aga_singular_params > 1:
            raise ValueError(
                "Exactly many of aga_product, aga_zip, or aga_params are True. "
                "Only 0 or 1 of the flags is allowed. \n"
                f"You got: "
                f"aga_product={aga_product}, aga_zip={aga_zip}, aga_params={aga_params}"
            )

        # pop aga keywords out
        aga_kwargs_dict = {
            kwd.value: kwargs.pop(kwd.value)
            for kwd in AgaReservedKeywords
            if kwd.value in kwargs
        }

        if aga_params:
            # build final params
            # So we're allowing [param(1, 2), [3, 4]] as a valid input
            self.final_params: List[_TestParam] = type(self).parse_params(
                *args, **kwargs
            )
        elif aga_singular_params:
            self.final_params = type(self).parse_singular_params(*args, **kwargs)
        elif aga_zip or aga_product:
            self.final_params = type(self).parse_zip_or_product(
                *args, aga_zip=aga_zip, aga_product=aga_product, **kwargs
            )
        else:
            self.final_params = type(self).parse_no_flag(*args, **kwargs)

        type(self).add_aga_kwargs(aga_kwargs_dict, self.final_params)

    @staticmethod
    def parse_params(*args: Iterable[Any], **kwargs: Any) -> List[_TestParam]:
        """Parse the parameters for param sequence."""
        if kwargs:
            raise ValueError("aga_params=True ignores non-aga kwargs")

        if len(args) != 1:
            raise ValueError(
                "aga_params=True requires exactly one iterable of sets of parameters"
            )

        return list(
            arg if isinstance(arg, _TestParam) else param(*arg) for arg in args[0]
        )

    @staticmethod
    def parse_singular_params(*args: Iterable[Any], **kwargs: Any) -> List[_TestParam]:
        """Parse the parameters for param sequence."""
        if kwargs:
            raise ValueError("aga_singular_params=True ignores non-aga kwargs")

        if len(args) != 1:
            raise ValueError(
                "aga_singular_params=True requires "
                "exactly one iterable of sets of parameters"
            )

        return list(
            arg if isinstance(arg, _TestParam) else param(arg) for arg in args[0]
        )

    @staticmethod
    def parse_zip_or_product(
        *args: Iterable[Any],
        aga_product: bool = False,
        aga_zip: bool = False,
        **kwargs: Any,
    ) -> List[_TestParam]:
        """Parse parameters for zip or product."""
        if not aga_zip ^ aga_product:
            raise ValueError("exactly one of aga_zip or aga_product must be True")
        combinator = product if aga_product else zip

        # ok so if the combinator is product
        # we are taking the cartesian product for all args and kwargs
        # and if the combinator is zip,
        # we are zipping all the args and kwargs, if there are any
        combined_args = list(combinator(*args))
        combined_kwargs = list(combinator(*kwargs.values()))

        # ======= validation checks =======
        if combinator is zip:
            # create empty args for zip if there are no args
            if combined_args and combined_kwargs:
                if len(combined_args) != len(combined_kwargs):
                    raise ValueError(
                        'length of "args" and "kwargs" must match in zip mode'
                    )
            elif combined_args:
                combined_kwargs = [()] * len(combined_args)
            elif combined_kwargs:
                combined_args = [()] * len(combined_kwargs)

        all_args_and_kwargs = list(combinator(combined_args, combined_kwargs))

        # ======= zipping all the args together =======
        return list(
            param(*curr_args, **dict(zip(kwargs.keys(), curr_kwargs)))
            for (curr_args, curr_kwargs) in all_args_and_kwargs
        )

    @staticmethod
    def parse_no_flag(*args: Iterable[Any], **kwargs: Any) -> List[_TestParam]:
        """Parse the parameters for no flag."""
        if kwargs:
            raise ValueError("`test_cases` with no flags ignores non-aga kwargs")

        return [param(arg) for arg in args]

    @staticmethod
    def add_aga_kwargs(
        aga_kwargs: Dict[str, Any], final_params: List[_TestParam]
    ) -> None:
        """Add aga_kwargs to the finalized parameters."""
        # process aga input type
        for aga_kwarg_key, aga_kwarg_value in aga_kwargs.items():
            if isinstance(aga_kwarg_value, Iterable) and not isinstance(
                aga_kwarg_value, str
            ):
                aga_kwargs[aga_kwarg_key] = list(aga_kwarg_value)
            else:
                aga_kwargs[aga_kwarg_key] = [aga_kwarg_value] * len(final_params)

        # validate aga input type
        if not all(
            len(aga_kwarg_value) == len(final_params)
            for aga_kwarg_value in aga_kwargs.values()
        ):
            # the length of the kwargs should be equal to the number of test cases
            # i.e. the length of the combined args
            raise ValueError(
                f"all aga_ keyword args must have the same length as the test cases, "
                f"which is {len(final_params)}"
            )

        # the aga kwargs list we want
        aga_kwargs_list: List[Dict[str, Any]] = [
            dict(zip(aga_kwargs.keys(), aga_kwarg_value))
            for aga_kwarg_value in zip(*aga_kwargs.values())
        ]

        if not aga_kwargs_list:
            # generate default aga kwargs dict if there are no aga kwargs
            aga_kwargs_list = [{} for _ in final_params]

        for final_param, kwargs in zip(final_params, aga_kwargs_list):
            final_param.update_aga_kwargs(**kwargs)

    def __call__(
        self, prob: Problem[ProblemParamSpec, ProblemOutputType]
    ) -> Problem[ProblemParamSpec, ProblemOutputType]:
        """Generate the test cases as a decorator."""
        for final_param in self.final_params:
            prob = final_param.generate_test_case(prob)

        return prob


test_cases = _TestParams  # pylint: disable=invalid-name
test_cases_params = partial(test_cases, aga_params=True)
test_cases_zip = partial(test_cases, aga_zip=True)
test_cases_product = partial(test_cases, aga_product=True)
test_cases_singular_params = partial(test_cases, aga_singular_params=True)
test_cases.params = test_cases_params
test_cases.product = test_cases_product
test_cases.zip = test_cases_zip
test_cases.singular_params = test_cases_singular_params



"""The core library functionality."""

from .parameter import (
    param,
    test_case,
    test_cases,
    test_cases_params,
    test_cases_product,
    test_cases_singular_params,
    test_cases_zip,
)
from .problem import Problem, group, problem
from .suite import AgaTestCase, AgaTestSuite, SubmissionMetadata, TestMetadata

__all__ = [
    "TestMetadata",
    "SubmissionMetadata",
    "AgaTestCase",
    "AgaTestSuite",
    "Problem",
    "problem",
    "group",
    "test_case",
    "test_cases",
    "param",
    "test_cases_params",
    "test_cases_zip",
    "test_cases_product",
    "test_cases_singular_params",
]



"""The core entrypoint for the gradescope frontend.

More complete documentation is in `__init__.py`.
"""

from importlib.abc import Traversable
from importlib.resources import files
from os import makedirs
from os.path import join as pathjoin
from shutil import copyfileobj
from sys import version_info
from tempfile import TemporaryDirectory
from typing import Iterable, Optional
from zipfile import ZipFile

from dill import dump  # type: ignore

from ..core import Problem
from ..core.problem import ProblemOutputType, ProblemParamSpec

# don't zip resources because we handle them manually
_ZIP_IGNORES = ("__pycache__", "resources")

# files we need present in the gradescope environment that aren't part of the `aga`
# source or of the specific problem
_GS_ENV_DEPS = ("setup.py", "run_autograder")

GS_UTILS_RESOURCE_DIR = "aga.gradescope.resources"


class InvalidProblem(BaseException):
    """The Problem failed some golden tests."""


def _handle_aga_zipping(zip_file: ZipFile, temp_working_dir_path: str) -> None:
    for src_path, resource_name in _copy_package_to(
        temp_working_dir_path, files("aga")
    ):
        zip_file.write(src_path, arcname=resource_name)


def _get_setup_shell_by_version() -> str:
    return f"setup-{version_info.major}{version_info.minor}.sh"


def _handle_gs_utils_zipping(zip_file: ZipFile, temp_working_dir_path: str) -> None:
    # copy the setup shell script according to the python version the user is using
    path = _manual_copy_resource_to(
        temp_working_dir_path, _get_setup_shell_by_version()
    )
    zip_file.write(path, arcname="setup.sh")

    # copy general gradescope utils
    for file in _GS_ENV_DEPS:
        path = _manual_copy_resource_to(temp_working_dir_path, file)
        zip_file.write(path, arcname=file)


def _handle_problem_zipping(
    zip_file: ZipFile,
    temp_working_dir: str,
    problem: Problem[ProblemParamSpec, ProblemOutputType],
) -> None:
    path = _dump_problem_into_dir(problem, temp_working_dir)
    zip_file.write(path, arcname="problem.pckl")


def into_gradescope_zip(
    problem: Problem[ProblemParamSpec, ProblemOutputType], path: Optional[str] = None
) -> str:
    """Convert a Problem into a gradescope autograder zip, returning its location.

    This is the high-level entrypoint for this module.
    """
    _check_problem(problem)
    zip_name = _get_zipfile_dest(path, problem)

    with TemporaryDirectory() as tempdir, ZipFile(zip_name, "w") as zip_f:
        _handle_aga_zipping(zip_f, tempdir)
        _handle_gs_utils_zipping(zip_f, tempdir)
        _handle_problem_zipping(zip_f, tempdir, problem)

    return zip_name


def _get_zipfile_dest(
    path: Optional[str], problem: Problem[ProblemParamSpec, ProblemOutputType]
) -> str:
    """Determine the destination in which to put the zip file.

    If `path` is none, this is the problem's name; otherwise it's just the provided
    path.
    """
    if path is None:
        return problem.name() + ".zip"
    else:
        return path


def _check_problem(problem: Problem[ProblemParamSpec, ProblemOutputType]) -> None:
    """Check whether `problem` is valid.

    Currently, this just runs the golden tests for problem.
    """
    try:
        problem.check()
    except AssertionError as err:
        raise InvalidProblem from err


def _copy_package_to(
    tempdir: str, package: Traversable, prefix: str = "aga"
) -> Iterable[tuple[str, str]]:
    """Recusrively copy each resource, yielding the file path and resource name."""
    # can we use `copytree` from shutil for this? or maybe `inspect.getsource`?

    makedirs(pathjoin(tempdir, prefix), exist_ok=True)
    # manually traverse the package structure, copying source files as we go
    for resource in package.iterdir():
        if resource.name not in _ZIP_IGNORES:
            if resource.is_dir():
                yield from _copy_package_to(
                    tempdir,
                    resource,
                    pathjoin(prefix, resource.name),
                )
            else:
                yield _copy_resource_to(tempdir, resource, prefix)


def _copy_resource_to(
    tempdir: str, resource: Traversable, prefix: str
) -> tuple[str, str]:
    """Copy the resource to tempdir, yielding the file path and resource name."""
    dest_path = pathjoin(tempdir, prefix, resource.name)
    with resource.open() as src:  # type: ignore
        with open(dest_path, "w", encoding="UTF-8") as dest:
            copyfileobj(src, dest)

    return (dest_path, pathjoin(prefix, resource.name))


def _manual_copy_resource_to(
    tempdir: str, fname: str, package: str = GS_UTILS_RESOURCE_DIR
) -> str:
    """Copy the resource at package.fname to tempdir/fname, returning the dest path."""
    dest_path = pathjoin(tempdir, fname)
    with files(package).joinpath(fname).open() as src:  # type: ignore
        with open(dest_path, "w", encoding="UTF-8") as dest:
            copyfileobj(src, dest)

    return dest_path


def _dump_problem_into_dir(
    problem: Problem[ProblemParamSpec, ProblemOutputType],
    tempdir: str,
    fname: str = "problem.pckl",
) -> str:
    """Dump a problem into a directory, returning the pckl file path."""
    path = pathjoin(tempdir, fname)
    _dump_problem_at_path(problem, path)
    return path


def _dump_problem_at_path(
    problem: Problem[ProblemParamSpec, ProblemOutputType], dest: str
) -> None:
    """Pickle the problem into a destination."""
    with open(dest, "wb") as file:
        dump(problem, file)



"""Test Suite for the aga.core module."""
from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass
from datetime import timedelta
from types import FunctionType
from typing import Any, Callable, Dict, Generic, Iterable, Sequence, Tuple, TypeVar
from unittest import TestCase, TestSuite
from unittest.mock import patch

from aga.core.environment import Environment

from ..config import AgaConfig, AgaTestConfig
from ..score import Prize, ScoredPrize, ScoreInfo, compute_scores
from ..util import text_diff
from .parameter import AgaKeywordContainer, _TestParam
from .utils import CaptureOut, Initializer, initializer

__all__ = ["TestMetadata", "SubmissionMetadata", "AgaTestCase", "AgaTestSuite"]

Output = TypeVar("Output")

_NEW_LINE = "\n"


@dataclass(frozen=True)
class TestMetadata:
    """Stores metadata about a specific test case."""

    max_score: float
    config: AgaTestConfig
    check_stdout: bool
    mock_input: bool
    hidden: bool = False


@dataclass(frozen=True)
class SubmissionMetadata:
    """Metadata for testing a submission, collected from the frontend.

    Attributes
    ----------
    total_score : float
        The problem's total score.
    time_since-due : timedelta
        The delta _from_ the due date _to_ the submission date, i.e. it's negative if
        the problem was submitted before the due date.
    previous_submissions : int
        The number of previous submissions.

    """

    total_score: float
    time_since_due: timedelta
    previous_submissions: int

    def is_on_time(self) -> bool:
        """Return true of the submission was on time."""
        return self.time_since_due <= timedelta()


class AgaTestCase(TestCase):
    """A `TestCase` which runs a single test of a `Problem`."""

    def __init__(
        self,
        test_input: "_TestInputs[Output]",
        golden: Callable[..., Output],
        under_test: Callable[..., Output],
        metadata: TestMetadata,
    ) -> None:
        super().__init__()
        self._test_input = test_input
        self._golden = golden
        self._under_test = under_test
        self._metadata = metadata

    @property
    def metadata(self) -> TestMetadata:
        """Get the test's metadata."""
        return self._metadata

    @property
    def test_input(self) -> _TestInputs[Output]:
        """Get the test input."""
        return self._test_input

    # pylint: disable=invalid-name
    # camelCase is required by unittest
    def runTest(self) -> None:
        """Run the test case."""
        self._test_input.check(self._golden, self._under_test, self._metadata)

    # pylint: disable=invalid-name
    # camelCase is required by unittest
    def shortDescription(self) -> str:
        """Dynamically generate the test name.

        This method is called by unittest.
        """
        return self.name

    @property
    def name(self) -> str:
        """Format the name of the test case."""
        return (
            self._test_input.aga_kwargs.name
            or self._metadata.config.name_fmt.format(
                args=self._test_input.param.args_repr(self._metadata.config.name_sep),
                kwargs=self._test_input.param.kwargs_repr(
                    self._metadata.config.name_sep
                ),
                sep=self._test_input.param.sep_repr(self._metadata.config.name_sep),
            )
        )

    @property
    def description(self) -> str | None:
        """Get the problem's description."""
        return self._test_input.aga_kwargs.description


class AgaTestSuite(TestSuite):
    """A thin wrapper around TestSuite that store a config."""

    def __init__(self, config: AgaConfig, tests: list[AgaTestCase]):
        super().__init__(tests)
        self.config = config


def generate_custom_input(input_list: Iterable[str]) -> Callable[[Any], str]:
    """Generate a custom input function for a test case."""
    _iterator = iter(input_list)

    def _custom_input(*args: Any) -> str:
        print(*args)
        return next(_iterator)

    return _custom_input


class _TestInputs(TestCase, Generic[Output]):
    """A single set of test inputs for a problem.

    These will be run against the golden solution and the function under test; their
    outputs will be compared, and a unittest failure raised if they differ.
    """

    # this tells unittest not to print their default assertion error messages
    longMessage = False

    def __init__(
        self,
        aga_param: _TestParam,
        mock_input: bool,
        env: Environment | None = None,
    ) -> None:
        super().__init__()
        self._mock_input = mock_input
        self.env = env
        self._param: _TestParam = aga_param
        self.score_info = ScoreInfo(
            self.aga_kwargs.weight, self.aga_kwargs.value, self.aga_kwargs.extra_credit
        )

    @property
    def args(self) -> Tuple[Any, ...]:
        """Get the positional arguments for the test case."""
        return self._param.args

    @property
    def kwargs(self) -> Dict[str, Any]:
        """Get the keyword arguments for the test case."""
        return self._param.kwargs

    @property
    def param(self) -> _TestParam:
        """Get the test parameter."""
        return self._param

    @property
    def aga_kwargs(self) -> AgaKeywordContainer:
        """Get the keyword arguments for the test case."""
        return self._param

    @property
    def mock_input(self) -> bool:
        """Get the mock input of the test case."""
        return self._mock_input

    @property
    def description(self) -> str | None:
        """Get the description of the test case."""
        return self.aga_kwargs.description

    @description.setter
    def description(self, desc: str | None) -> None:
        """Set the description of the test case."""
        self.aga_kwargs.description = desc

    @property
    def name(self) -> str | None:
        """Get the description of the test case."""
        return self.aga_kwargs.name

    @name.setter
    def name(self, name: str | None) -> None:
        """Set the description of the test case."""
        self.aga_kwargs.name = name

    def _eval_mock_input(
        self, answer: Callable[..., Output], check_output: bool = False
    ) -> Tuple[str | None, Output]:
        """Evaluate func on the arguments."""
        # deepcopy in case the student submission mutates arguments; we don't want it to
        # mess with our copy of the arguments
        with CaptureOut(check_output) as stdout, patch(
            "builtins.input", generate_custom_input(deepcopy(self.args))
        ):
            result = answer()

        return stdout.value, result

    def _test_mock_input(
        self,
        golden: Callable[..., Output],
        under_test: Callable[..., Output],
        eq_fn: Callable[[_TestInputs[Output], Output, Output, TestMetadata, str], bool],
        metadata: TestMetadata,
    ) -> None:
        golden_stdout, golden_result = self._eval_mock_input(
            golden, metadata.check_stdout
        )

        under_test_stdout, under_test_result = self._eval_mock_input(
            under_test, metadata.check_stdout
        )

        eq_fn(
            self,
            golden_result,
            under_test_result,
            metadata,
            metadata.config.failure_msg,
        )

        self._assert_output_eq(golden_stdout, under_test_stdout, metadata)

    def _eval_regular(
        self, answer: Callable[..., Output], capture_output: bool
    ) -> Tuple[str | None, Output]:
        """Evaluate func on the arguments."""
        with CaptureOut(capture_output) as stdout:
            result = answer(*deepcopy(self.args), **deepcopy(self.kwargs))

        return stdout.value, result

    def _test_regular(
        self,
        golden: Callable[..., Output],
        under_test: Callable[..., Output],
        eq_fn: Callable[[_TestInputs[Output], Output, Output, TestMetadata, str], bool],
        metadata: TestMetadata,
    ) -> None:
        golden_stdout, golden_result = self._eval_regular(golden, metadata.check_stdout)
        under_test_stdout, under_test_result = self._eval_regular(
            under_test, metadata.check_stdout
        )

        eq_fn(
            self,
            golden_result,
            under_test_result,
            metadata,
            metadata.config.failure_msg,
        )

        self._assert_output_eq(golden_stdout, under_test_stdout, metadata)

    def _eval_pipeline(
        self, answer: Callable[..., Output], check_output: bool
    ) -> Tuple[str | None, Sequence[Any]]:
        self.description = ""
        temp_name = initializer.DEFAULT_NEW_INSTANCE_NAME
        if isinstance(answer, (type, FunctionType)):
            temp_name = answer.__name__

        with CaptureOut(check_output) as stdout:
            results = [None]

            if len(self.args) > 0:
                pipeline_processes = self.args

                for process in pipeline_processes:
                    if isinstance(process, Initializer):
                        answer = process(answer)
                        results.append(None)
                        self.description += (
                            f"{process.DEFAULT_NEW_INSTANCE_NAME} = "
                            f"{temp_name}{process}\n"
                        )
                        temp_name = process.DEFAULT_NEW_INSTANCE_NAME
                    else:
                        results.append(process(answer, results[-1]))
                        self.description += f"{temp_name}{process}\n"

        return stdout.value, results

    def _test_pipeline(
        self,
        golden: Callable[..., Output],
        under_test: Callable[..., Output],
        eq_fn: Callable[[_TestInputs[Output], Output, Output, TestMetadata, str], bool],
        metadata: TestMetadata,
    ) -> None:
        golden_stdout, golden_result = self._eval_pipeline(
            golden, metadata.check_stdout
        )
        under_test_stdout, under_test_result = self._eval_pipeline(
            under_test, metadata.check_stdout
        )

        if len(under_test_result) != len(golden_result):
            raise AssertionError(
                f"Expected {len(golden_result)} results, "
                f"but got {len(under_test_result)}"
            )

        for i, (golden_res, under_test_res) in enumerate(
            zip(golden_result[1:], under_test_result[1:])
        ):
            eq_fn(
                self,
                golden_res,
                under_test_res,
                metadata,
                f"The first diff happens in {i}th process (`{self.args[i]}`) "
                f"of the pipeline.\n"
                "The output is {output} and the expected is {expected}\n"
                f"Everything before the {i}th process (`{self.args[i]}`) "
                f"of the pipeline is correct:\n"
                f"{_NEW_LINE.join(map(str, self.args[:i + 1]))}\n\n"
                f"{metadata.config.failure_msg}",
            )

        self._assert_output_eq(golden_stdout, under_test_stdout, metadata)

    def _pick_test_fn(
        self,
    ) -> Callable[
        [
            Callable[..., Output],
            Callable[..., Output],
            Callable[[_TestInputs[Output], Output, Output, TestMetadata, str], bool],
            TestMetadata,
        ],
        None,
    ]:
        if self.mock_input:
            test_fn = self._test_mock_input
        elif self.aga_kwargs.is_pipeline:
            test_fn = self._test_pipeline
        else:
            test_fn = self._test_regular

        return test_fn

    def _pick_eval_fn(
        self,
    ) -> Callable[[Callable[..., Output], bool], Tuple[str, Any]]:
        if self.mock_input:
            test_fn: Any = self._eval_mock_input
        elif self.aga_kwargs.is_pipeline:
            test_fn = self._eval_pipeline
        else:
            test_fn = self._eval_regular

        return test_fn  # type: ignore

    def check(
        self,
        golden: Callable[..., Output],
        under_test: Callable[..., Output],
        metadata: TestMetadata,
    ) -> None:
        """Assert that `golden` and `under_test` give the same output.

        This method relies on `TestCase.assertEqual`, which will raise a unittest error
        if they differ. This means it can be plugged into any unittest TestCase as a
        helper method.
        """
        if self.aga_kwargs.override_test:
            self.aga_kwargs.override_test(self, golden, under_test)
            return

        if self.aga_kwargs.override_check:
            check = self.aga_kwargs.override_check
        else:
            check = type(self)._assert_eq

        test_fn = self._pick_test_fn()

        test_fn(golden, under_test, check, metadata)

    def _assert_output_eq(
        self, expected: str | None, got: str | None, metadata: TestMetadata
    ) -> None:
        self._assert_eq(
            expected or "",
            got or "",
            metadata,
            metadata.config.stdout_differ_msg,
        )

    def _assert_eq(
        self,
        expected: Any,
        got: Any,
        metadata: TestMetadata,
        msg_format: str,
    ) -> None:
        """Assert that expected equals got, formatting `msg_format` if not."""
        # we can only diff strings
        if isinstance(expected, str) and isinstance(got, str):
            diff_explanation = metadata.config.diff_explanation_msg
            diff = text_diff(got, expected)
        else:
            diff_explanation = ""
            diff = ""

        if isinstance(expected, float) and isinstance(got, (float, int)):
            comparator = self.assertAlmostEqual  # type: ignore

        else:
            comparator = self.assertEqual  # type: ignore

        comparator(  # type: ignore
            got,
            expected,
            msg=msg_format.format(
                input=repr(self),
                expected=repr(expected),
                output=repr(got),
                diff_explanation=diff_explanation,
                diff=diff,
            ),
        )

    def generate_test_case(
        self,
        golden: Callable[..., Output],
        under_test: Callable[..., Output],
        score: float,
        config: AgaConfig,
    ) -> AgaTestCase:
        """Generate a TestCase which tests `golden` against `under_test`."""
        metadata = TestMetadata(
            hidden=self.aga_kwargs.hidden,
            config=config.test,
            max_score=score,
            check_stdout=config.problem.check_stdout,
            mock_input=config.problem.mock_input,
        )
        return AgaTestCase(self, golden, under_test, metadata)

    def __repr__(self) -> str:
        """Get a string representation of the test case."""
        args_repr = self._param.args_repr(",")
        kwargs_repr = self._param.kwargs_repr(",")
        sep = self._param.sep_repr(",")

        return args_repr + sep + kwargs_repr

    def check_one(self, golden: Callable[..., Output]) -> None:
        """Check that the golden solution is correct."""
        if (
            self.aga_kwargs.expect is not None
            or self.aga_kwargs.expect_stdout is not None
        ):
            if self.aga_kwargs.override_test:

                def dummy_tester(*_: Any, **__: Any) -> Output:
                    # https://github.com/python/mypy/issues/4805 ehh
                    return self.aga_kwargs.expect  # type: ignore

                try:
                    self.aga_kwargs.override_test(
                        self,
                        dummy_tester,
                        golden,
                    )
                except AssertionError as e:
                    raise AssertionError(
                        "Your solution doesn't pass the overridden test. \n"
                        "Test parameters: \n"
                        f"{self._param}"
                    ) from e
            else:
                self._default_check(golden)

    def _default_check(self, golden: Callable[..., Output]) -> None:
        """Check that the golden solution is correct."""
        eval_fn = self._pick_eval_fn()
        golden_stdout, golden_output = eval_fn(golden, True)

        # compare output
        if self.aga_kwargs.expect is not None:
            compared_expect_out = self.aga_kwargs.expect
            if self.aga_kwargs.is_pipeline:
                golden_output = golden_output[1:]
                compared_expect_out = list(self.aga_kwargs.expect)

            # if check is overridden, use it
            if self.aga_kwargs.override_check is not None:
                try:
                    self.aga_kwargs.override_check(
                        self, compared_expect_out, golden_output
                    )
                except AssertionError as e:
                    raise AssertionError(
                        "Your solution doesn't pass the overridden check. \n"
                        "Test parameters: \n"
                        f"{self._param}"
                    ) from e
            else:
                self.assertEqual(golden_output, self.aga_kwargs.expect)

        # compare stdout
        if self.aga_kwargs.expect_stdout is not None:
            if isinstance(self.aga_kwargs.expect_stdout, str):
                compared_golden_stdout = golden_stdout
                compared_expected_stdout = self.aga_kwargs.expect_stdout
            elif isinstance(self.aga_kwargs.expect_stdout, Sequence) and isinstance(
                golden_stdout, str
            ):
                compared_golden_stdout = golden_stdout.splitlines()
                compared_expected_stdout = list(self.aga_kwargs.expect_stdout)
            else:
                raise TypeError(
                    "expect_stdout must be a string or a sequence of strings"
                )

            self.assertEqual(compared_expected_stdout, compared_golden_stdout)


class _TestInputGroup(Generic[Output]):
    """A group of test cases with shared configuration."""

    def __init__(
        self, weight: int = 1, value: float = 0.0, extra_credit: float = 0.0
    ) -> None:
        self._test_cases: list[_TestInputs[Output]] = []
        self._prizes: list[Prize] = []
        self.score_info = ScoreInfo(weight, value, extra_credit)

    def add_test_case(self, case: _TestInputs[Output]) -> None:
        """Add a test case to the group."""
        self._test_cases.append(case)

    def add_prize(self, prize: Prize) -> None:
        """Add a prize to the group."""
        self._prizes.append(prize)

    def generate_test_suite(
        self,
        golden: Callable[..., Output],
        under_test: Callable[..., Output],
        group_score: float,
        config: AgaConfig,
    ) -> tuple[AgaTestSuite, list[ScoredPrize]]:
        """Generate a test suite from all the test cases for this group."""
        suite = AgaTestSuite(config, [])

        score_infos = [case.score_info for case in self._test_cases] + [
            prize.score_info for prize in self._prizes
        ]
        scores = compute_scores(score_infos, group_score)

        for score, case in zip(
            scores, self._test_cases
        ):  # type: float, _TestInputs[Output]
            suite.addTest(case.generate_test_case(golden, under_test, score, config))

        scored_prizes = []
        for score, prize in zip(reversed(scores), reversed(self._prizes)):
            # reverse scores so we get the ones that correspond to the prizes
            scored_prizes.append(
                ScoredPrize(prize=prize, max_score=score)  # type: ignore
            )

        return suite, scored_prizes

    def check_one(self, golden: Callable[..., Output]) -> None:
        """Check the golden solution against all test cases."""
        for case in self._test_cases:
            case.check_one(golden)



"""Problem and utilities."""
from __future__ import annotations

from types import ModuleType
from typing import (
    TYPE_CHECKING,
    Callable,
    Generic,
    Optional,
    ParamSpec,
    TypeVar,
    Iterable,
)

from ..config import AgaConfig
from ..score import Prize, ScoredPrize, compute_scores
from .environment import Environment
from .suite import AgaTestSuite, SubmissionMetadata, _TestInputGroup, _TestInputs

# pylint: disable=invalid-name
ProblemParamSpec = ParamSpec("ProblemParamSpec")
ProblemOutputType = TypeVar("ProblemOutputType")


if TYPE_CHECKING:
    from .parameter import _TestParam

__all__ = ["Problem", "problem", "group", "ProblemOutputType", "ProblemParamSpec"]


# pylint: disable=R0902
class Problem(Generic[ProblemParamSpec, ProblemOutputType]):
    """Stores tests for a single problem."""

    # pylint: disable=R0913
    def __init__(
        self,
        golden: Callable[ProblemParamSpec, ProblemOutputType],
        name: str,
        config: AgaConfig,
        is_script: bool,
        env_targets: Iterable[str] = (),
    ) -> None:
        self._golden: Callable[ProblemParamSpec, ProblemOutputType] = golden
        self._name = name
        self._config = config
        self._ungrouped_prizes: list[Prize] = []
        self._ungrouped_tests: list[_TestInputs[ProblemOutputType]] = []
        self._groups: list[_TestInputGroup[ProblemOutputType]] = []
        self._environment: Environment = Environment(env_targets)
        self.is_script = is_script

    def add_test_case(self, param: _TestParam) -> None:
        """Add a test case to the current group.

        Student solutions will be checked against the golden solution; i.e., this method
        does _not_ produce a test of the golden solution.
        """
        case: _TestInputs[ProblemOutputType] = _TestInputs(
            param, mock_input=self._config.problem.mock_input, env=self.environment
        )
        self._ungrouped_tests.append(case)

    def add_prize(self, prize: Prize) -> None:
        """Add a prize to the current group."""
        self._ungrouped_prizes.append(prize)

    def add_group(self, grp: _TestInputGroup[ProblemOutputType]) -> None:
        """Add a group to the problem."""
        for case in self._ungrouped_tests:
            grp.add_test_case(case)

        for prize in self._ungrouped_prizes:
            grp.add_prize(prize)

        self._groups.append(grp)
        self._ungrouped_tests = []
        self._ungrouped_prizes = []

    def config(self) -> AgaConfig:
        """Get access to the problem's config."""
        return self._config

    def check(self) -> None:
        """Check that the problem is correct.

        Currently, this runs all tests of the golden solution.
        """
        for grp in self._virtual_groups():
            grp.check_one(self._golden)

    def generate_test_suite(
        self,
        under_test: Callable[ProblemParamSpec, ProblemOutputType],
        metadata: SubmissionMetadata,
    ) -> tuple[AgaTestSuite, list[ScoredPrize]]:
        """Generate a `TestSuite` for the student submitted function.

        Neither the generated test suite nor the body of this function will run golden
        tests; instead, golden test cases are treated as equivalent to ordinary ones. To
        test the golden function, `check` should be used instead.

        Parameters
        ----------
        under_test : Callable[ProblemParamSpec, ProblemOutputType]
            The student submitted function.
        metadata : SubmissionMetadata
            The submission metadata.

        Returns
        -------
        AgaTestSuite
            A unittest test suite containing one test for each TestInput in this
            problem, checking the result of the problem's golden solution against
            `under_test`.
        list[ScorePrize]
            The prizes for the problem, with scores assigned.
        """
        ret_suite = AgaTestSuite(self._config, [])

        groups = self._virtual_groups()

        score_infos = [grp.score_info for grp in groups]
        scores = compute_scores(score_infos, metadata.total_score)

        ret_prizes = []
        for score, grp in zip(scores, groups):
            suite, prizes = grp.generate_test_suite(
                self._golden, under_test, score, self._config
            )
            ret_suite.addTest(suite)
            ret_prizes += prizes

        return ret_suite, ret_prizes

    @property
    def golden(self) -> Callable[ProblemParamSpec, ProblemOutputType]:
        """The gold solution property."""
        return self._golden

    def name(self) -> str:
        """Get the problem's name."""
        return self._name

    @property
    def environment(self) -> Environment:
        """The environment values captured from the problem module."""
        return self._environment

    def update_env_from(
        self, mod: ModuleType
    ) -> Problem[ProblemParamSpec, ProblemOutputType]:
        """Update the environment values from a given module."""
        self.environment.update_from_module(mod)
        return self

    def expected_symbol(self) -> str:
        """Get the name of the symbol that should be tested against."""
        return self._golden.__name__

    def update_config_weak(self, config: AgaConfig) -> None:
        """Update any non-default items in self.config."""
        self._config.update_weak(config)

    def _virtual_groups(self) -> list[_TestInputGroup[ProblemOutputType]]:
        """Get the list of groups, plus the current group under construction.

        We need to do it this way because while the problem is being read we don't know
        the configuration of the last test group yet.
        """
        if self._ungrouped_tests or self._ungrouped_prizes:
            virtual_group: _TestInputGroup[ProblemOutputType] = _TestInputGroup()

            for case in self._ungrouped_tests:
                virtual_group.add_test_case(case)

            for prize in self._ungrouped_prizes:
                virtual_group.add_prize(prize)

            return self._groups + [virtual_group]

        else:
            return self._groups

    def __call__(
        self, *args: ProblemParamSpec.args, **kwargs: ProblemParamSpec.kwargs
    ) -> ProblemOutputType:  # type: ignore
        """Enable the ability to call the golden solution as if the problem were it."""
        return self._golden(*args, **kwargs)


def problem(
    name: Optional[str] = None,
    script: bool = False,
    check_stdout: Optional[bool] = None,
    mock_input: Optional[bool] = None,
    envs: Iterable[str] = (),
) -> Callable[
    [Callable[ProblemParamSpec, ProblemOutputType]],
    Problem[ProblemParamSpec, ProblemOutputType],
]:
    """Declare a function as the golden solution to a problem.

    This method should decorate a function which is known to produce the correct
    outputs, which we refer to as the "golden solution". It also provides facilities
    for testing that solution, via golden test cases, constructed by passing the
    output argument to the test_case decorator.

    Parameters
    ----------
    name : Optional[str]
        The problem's name. If None (the default), the wrapped function's name will be
        used.
    script : bool
        Whether the problem represents a script, as opposed to a function. Implies
        `check_stdout` and `mock_input` unless they are passed explicitly.
    check_stdout : Optional[bool]
        Overrides the `problem.check_stdout` configuration option. If True, check the
        golden solution's stdout against the student submission's for all test cases.
    mock_input : Optional[bool]
        Overrides the `problem.mock_input` configuration option. If True, test cases for
        this problem will be interpreted as mocked outputs of `builtins.input`, rather
        than inputs to the function.
    envs : Iterable[str]
        The environment values to be captured

    Returns
    -------
    Callable[[Callable[ProblemInput, T]], Problem[T]]
        A decorator which turns a golden solution into a problem.
    """
    config = AgaConfig()

    if check_stdout is not None:
        config.problem.check_stdout = check_stdout
        config.problem.check_stdout_overridden = True

    if mock_input is not None:
        config.problem.mock_input = mock_input
        config.problem.mock_input_overridden = True

    def outer(
        func: Callable[ProblemParamSpec, ProblemOutputType]
    ) -> Problem[ProblemParamSpec, ProblemOutputType]:
        problem_name = name or func.__name__

        if script:
            if check_stdout is None:
                config.problem.check_stdout = True
                config.problem.check_stdout_overridden = True

            if mock_input is None:
                config.problem.mock_input = True
                config.problem.mock_input_overridden = True

        return Problem(func, problem_name, config, script, env_targets=envs)

    return outer


def group(
    weight: int = 1,
    value: float = 0.0,
    extra_credit: float = 0.0,
) -> Callable[
    [Problem[ProblemParamSpec, ProblemOutputType]],
    Problem[ProblemParamSpec, ProblemOutputType],
]:
    """Declare a group of problems.

    Parameters
    ----------
    weight : int
        The group's relative weight to the problem's score. See :ref:`Determining Score`
        for details.
    value : float
        The group's absolute score. See :ref:`Determining Score` for details.
    extra_credit : float
        The group's extra credit. See :ref:`Determining Score` for details.

    Returns
    -------
    Callable[[Problem[T]], Problem[T]]
        A decorator which adds the group to a problem.
    """

    def outer(
        prob: Problem[ProblemParamSpec, ProblemOutputType]
    ) -> Problem[ProblemParamSpec, ProblemOutputType]:
        prob.add_group(_TestInputGroup(weight, value, extra_credit))
        return prob

    return outer



"""Environment value wrapper."""
from __future__ import annotations

import types
from typing import Iterable, Any


class Environment(dict[str, Any]):
    """Environment value wrapper."""

    def __init__(self, env_targets: Iterable[str]) -> None:
        super().__init__()
        for target in env_targets:
            self[target] = None

    def update_from_module(self, mod: types.ModuleType) -> None:
        """Update environment values from a given module."""
        for value_name in self.keys():
            if value_name not in vars(mod):
                raise ValueError(
                    f"The variable `{value_name} does not exist in the module {mod}"
                )
            self[value_name] = getattr(mod, value_name)

    def __getattr__(self, item: str) -> Any:
        """Get the value of an environment variable."""
        try:
            return self[item]
        except KeyError as e:
            raise AttributeError(e) from e
